from __future__ import unicode_literals

from macaroons.raw_macaroon import RawMacaroon
from macaroons.exceptions import MacaroonInitException
from macaroons.utils import convert_to_bytes, convert_to_string


class Macaroon:

    def __init__(self,
                 location=None,
                 identifier=None,
                 key=None,
                 serialized=None):

        if location and identifier and key:
            self._raw_macaroon = RawMacaroon(
                location=convert_to_bytes(location),
                identifier=convert_to_bytes(identifier),
                key=convert_to_bytes(key)
            )
        elif serialized:
            self._raw_macaroon = RawMacaroon(
                serialized=convert_to_bytes(serialized)
            )
        else:
            raise MacaroonInitException(
                'Must supply either: (location, id, key) or (serialized).'
            )

    @property
    def location(self):
        return convert_to_string(self._raw_macaroon.location)

    @property
    def identifier(self):
        return convert_to_string(self._raw_macaroon.identifier)

    @property
    def signature(self):
        return convert_to_string(self._raw_macaroon.signature)

    @property
    def caveats(self):
        return self._raw_macaroon.caveats

    def validate(self):
        return self._raw_macaroon.validate()

    def copy(self):
        return Macaroon(serialized=self.serialize())

    # Concatenates location, id, all caveats, and signature,
    # and then base64 encodes them
    def serialize(self):
        return self._raw_macaroon.serialize().decode('ascii')

    def serialize_json(self):
        return self._raw_macaroon.serialize_json()

    def inspect(self):
        return self._raw_macaroon.inspect()

    def is_same(self, macaroon):
        return self._raw_macaroon.is_same(macaroon._raw_macaroon)

    def first_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.first_party()]

    def third_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.third_party()]

    # Protects discharge macaroons in the event they are sent to
    # the wrong location by binding to the root macaroon
    def prepare_for_request(self, macaroon):
        protected = self.copy()
        protected._raw_macaroon = self._raw_macaroon.prepare_for_request(macaroon._raw_macaroon)
        return protected

    # The existing macaroon signature is the key for hashing the
    # caveat being added. This new hash becomes the signature of
    # the macaroon with caveat added.
    def add_first_party_caveat(self, predicate):
        self._raw_macaroon.add_first_party_caveat(
            convert_to_bytes(predicate)
        )
        return self

    # The third party caveat key is encrypted useing the current signature, and
    # the caveat is added to the list. The existing macaroon signature
    # is the key for hashing the string (verificationId + caveatId).
    # This new hash becomes the signature of the macaroon with caveat added.
    def add_third_party_caveat(self, location, key, key_id):
        return self._raw_macaroon.add_third_party_caveat(
            convert_to_bytes(location),
            convert_to_bytes(key),
            convert_to_bytes(key_id)
        )



import hmac
from base64 import standard_b64decode

from libnacl.secret import SecretBox
from libnacl import crypto_secretbox_NONCEBYTES
from streql import equals

from macaroons.macaroon import Macaroon
from macaroons.raw_macaroon import RawMacaroon
from macaroons.exceptions import (MacaroonInvalidSignatureException,
                                  MacaroonUnmetCaveatException)
from macaroons.utils import (convert_to_bytes,
                             convert_to_string,
                             truncate_or_pad)


class Verifier:

    predicates = []
    callbacks = []

    def satisfy_exact(self, predicate):
        if predicate is None:
            raise TypeError('Predicate cannot be none.')
        self.predicates.append(convert_to_string(predicate))

    def satisfy_general(self, func):
        if not hasattr(func, '__call__'):
            raise TypeError('General caveat verifiers must be functions.')
        self.callbacks.append(func)

    def verify(self, macaroon, key, discharge_macaroons=None):
        compare_macaroon = RawMacaroon(
            location=convert_to_bytes(macaroon.location),
            identifier=convert_to_bytes(macaroon.identifier),
            key=convert_to_bytes(key)
        )

        self._verify_caveats(
            macaroon._raw_macaroon,
            compare_macaroon,
            discharge_macaroons
        )

        if not self._signatures_match(
                macaroon._raw_macaroon, compare_macaroon):
            raise MacaroonInvalidSignatureException('Signatures do not match.')

        return True

    # Discharge macaroons are bound to the root,
    # extra steps to calculate the signature
    def verify_discharge(self, root, macaroon, key, discharge_macaroons=None):
        compare_macaroon = RawMacaroon(
            location=convert_to_bytes(macaroon.location),
            identifier=convert_to_bytes(macaroon.identifier),
            signature=macaroon._raw_macaroon._macaroon_hmac(
                convert_to_bytes(key),
                convert_to_bytes(macaroon.identifier)
            )
        )

        self._verify_caveats(macaroon, compare_macaroon, discharge_macaroons)

        compare_macaroon = root.prepare_for_request(compare_macaroon)

        if not self._signatures_match(macaroon, compare_macaroon):
            raise MacaroonInvalidSignatureException('Signatures do not match.')

        return True

    def _verify_caveats(self, macaroon, compare_macaroon, discharge_macaroons):
        for caveat in macaroon.caveats:
            if caveat.first_party():
                caveatMet = self._verify_first_party_caveat(
                    caveat,
                    compare_macaroon
                )
            else:
                caveatMet = self._verify_third_party_caveat(
                    caveat,
                    macaroon,
                    compare_macaroon,
                    discharge_macaroons
                )

            if not caveatMet:
                raise MacaroonUnmetCaveatException(
                    'Caveat not met. Unable to satisify: ' + caveat.caveatId
                )

    def _verify_first_party_caveat(self, caveat, compare_macaroon):
        caveatMet = False
        if caveat.caveatId in self.predicates:
            caveatMet = True
        else:
            for callback in self.callbacks:
                if callback(caveat.caveatId):
                    caveatMet = True
        if caveatMet:
            compare_macaroon.add_first_party_caveat(caveat._caveatId)
        return caveatMet

    def _verify_third_party_caveat(self,
                                   caveat,
                                   root_macaroon,
                                   compare_macaroon,
                                   discharge_macaroons):
        caveatMet = False

        caveat_macaroon = \
            next((m for m in discharge_macaroons
                  if m.identifier == caveat.caveatId), None)

        if not caveat_macaroon:
            raise MacaroonUnmetCaveatException(
                'Caveat not met. No discharge macaroon found for identifier: '
                + caveat.caveatId
            )

        caveat_key, nonce = self._extract_caveat_key(compare_macaroon, caveat)

        caveat_macaroon_verifier = Verifier()
        caveat_macaroon_verifier.predicates = self.predicates
        caveat_macaroon_verifier.callbacks = self.callbacks

        caveatMet = caveat_macaroon_verifier.verify_discharge(
            root_macaroon,
            caveat_macaroon,
            caveat_key,
            discharge_macaroons
        )
        if caveatMet:
            compare_macaroon._add_third_party_caveat_direct(
                caveat._location,
                caveat_key,
                caveat._caveatId,
                nonce=nonce
            )
        return caveatMet

    def _extract_caveat_key(self, compare_macaroon, caveat):
        key = truncate_or_pad(compare_macaroon.signature)
        box = SecretBox(key=key)
        decoded_vid = standard_b64decode(caveat.verificationKeyId)
        nonce = decoded_vid[:crypto_secretbox_NONCEBYTES]
        decrypted = box.decrypt(decoded_vid)
        return decrypted, nonce

    def _signatures_match(self, m1, m2):
        # uses a constant-time compare
        return equals(m1.signature, m2.signature)



import os

from setuptools import find_packages, setup

from macaroons import __version__


# Get long_description from index.rst:
here = os.path.dirname(os.path.abspath(__file__))
f = open(os.path.join(here, 'docs', 'index.rst'))
long_description = f.read().strip()
long_description = long_description.split('split here', 1)[1]
f.close()

setup(
    name='pymacaroons',
    version=__version__,
    description='Macaroon library for Python',
    author='Evan Cordell',
    author_email='evan.cordell@localmed.com',
    url='',
    license='MIT',
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    long_description=long_description,
    install_requires=[
        'six>=1.7.3',
        'libnacl>=1.3.5',
        'streql>=3.0.2',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security :: Cryptography',
        'Topic :: Security'
    ],
)
